using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Threading;
using unoidl.com.sun.star.awt;
using unoidl.com.sun.star.beans;
using unoidl.com.sun.star.bridge;
using unoidl.com.sun.star.container;
using unoidl.com.sun.star.drawing;
using unoidl.com.sun.star.frame;
using unoidl.com.sun.star.graphic;
using unoidl.com.sun.star.lang;
using unoidl.com.sun.star.sheet;
using unoidl.com.sun.star.table;
using unoidl.com.sun.star.text;
using unoidl.com.sun.star.uno;
using unoidl.com.sun.star.util;

// addpix
// add, resize, and center pictures in condo inventory spreadsheet

class Addpix {

  // contents of items from spreadsheet rows
  public class Item {
    public int nrow;
    public string pqitem;
    public string pqbox;
    public string pqdesc;
    public string pqJRC;
    public string pqECH;
    public string pqJCC;
    public string pqval;
    public string pqebay;
    public string pqpic1;
    public string pqpic2;
    public string pqpic3;
    public string pqpic4;
  };

  // contents of shapes containing graphic images
  public class Img {
    public XShape XS;
    public int nrow;
    public int ncol;
    public string pqnam;
    public unoidl.com.sun.star.awt.Size Szs =  // shape (current) size
      new unoidl.com.sun.star.awt.Size(0,0);
    public unoidl.com.sun.star.awt.Point Pts = // shape (current) position
      new unoidl.com.sun.star.awt.Point(0,0);
    public unoidl.com.sun.star.awt.Size Szg =  // jpg pixel size
      new unoidl.com.sun.star.awt.Size(0,0);
    public unoidl.com.sun.star.awt.Size Szc =  // cell size
      new unoidl.com.sun.star.awt.Size(0,0);
    public unoidl.com.sun.star.awt.Point Ptc = // cell position
      new unoidl.com.sun.star.awt.Point(0,0);
    public unoidl.com.sun.star.awt.Size Szn =  // shape (new) size
      new unoidl.com.sun.star.awt.Size(0,0);
    public unoidl.com.sun.star.awt.Point Ptn = // shape (new) position
      new unoidl.com.sun.star.awt.Point(0,0);
    public bool lfile; // jpg file exists
    public bool lshap; // shape exists
  };

  [STAThread]
  static void Main(string[] ppqargs) {

    // default paths and names (my spreadsheet)
    // string pqssh = "D:\\Documents\\Personal\\Mom & Dad\\EMC Estate\\"+
    //                "Inventory\\EMC Condo Contents Inventory.ods";
    // string pqimg = "D:\\Documents\\Personal\\Mom & Dad\\EMC Estate\\"+
    //                "Inventory\\Images\\";
    // string pqsnm = "Condo_Items";

    // default paths and names (sample spreadsheet)
    string pqssh = "C:\\ods-addpix\\Sample_Items.ods";
    string pqimg = "C:\\ods-addpix\\Images\\";
    string pqsnm = "Sheet1";

    // process argument list
    bool lok = true;
    bool ldel = false;
    bool lrfr = false;
    bool lvrb = false;
    bool lro = false;
    for (int narg = 0; narg < ppqargs.Length; narg++) {
      string pqarg = ppqargs[narg].ToLower();
      if (pqarg.Length == 4 && pqarg == "-del") ldel = true;
      else if (pqarg.Length == 4 && pqarg == "-rfr") lrfr = true;
      else if (pqarg.Length == 4 && pqarg == "-vrb") lvrb = true;
      else if (pqarg.Length == 3 && pqarg == "-ro") lro = true;
      else if (pqarg.Length == 4 && pqarg == "-ssh" &&
               narg+1 < ppqargs.Length) {
        pqssh = ppqargs[narg+1];
        if (!File.Exists(pqssh)) {
          Console.WriteLine("File "+pqssh+" does not exist");
          lok = false;
        }
        narg++;
      }
      else if (pqarg.Length == 4 && pqarg == "-img" &&
               narg+1 < ppqargs.Length) {
        pqimg = ppqargs[narg+1].TrimEnd('\\');
        if (!Directory.Exists(pqimg)) {
          Console.WriteLine("Folder "+pqimg+" does not exist");
          lok = false;
        }
        pqimg = pqimg+'\\';
        narg++;
      }
      else if (pqarg.Length == 4 && pqarg == "-snm" &&
               narg+1 < ppqargs.Length) {
        pqsnm = ppqargs[narg+1];
        narg++;
      }
      else lok = false;
    }
    if (ldel && lrfr) lok = false;

    // print instructions for bad options
    if (!lok) {
      string pqexepth = Process.GetCurrentProcess().MainModule.FileName;
      System.DateTime DTmod = File.GetLastWriteTime(pqexepth);
      Console.Error.WriteLine();
      Console.Error.WriteLine("ADDPIX - add images to spreadsheet - "+
                              DTmod.ToString("MM/dd/yyyy"));
      Console.Error.WriteLine();
      Console.Error.WriteLine("To execute: ADDPIX pars");
      Console.Error.WriteLine("   where 'pars' is one or more optional "+
                              "parameters from the following list:");
      Console.Error.WriteLine();
      Console.Error.WriteLine("   -del  delete all images from spreadsheet");
      Console.Error.WriteLine("   -rfr  refresh all images in spreadsheet");
      Console.Error.WriteLine("         (delete all spreadsheet images and "+
                              "reinsert them from image files)");
      Console.Error.WriteLine("   -vrb  verbose output");
      Console.Error.WriteLine("   -ro   read only (no changes made)");
      Console.Error.WriteLine("   -ssh  path to spreadsheet file");
      Console.Error.WriteLine("   -img  path to image file folder");
      Console.Error.WriteLine("   -snm  spreadsheet sheet name");
      Console.Error.WriteLine();
      Console.Error.WriteLine("   -ssh, -img, and -snm require another "+
                              "following parameter specifying the text");
      Console.Error.WriteLine();
      Console.Error.WriteLine("   -del and -rfr cannot be used together");
      return;
    }

    // set up some stuff
    bool lreadonly = lro;
    bool lchanged = false;

/*  begin: code to open a spreadsheet for reading and writing */

    // get the desktop object
    XComponentContext XCC = uno.util.Bootstrap.bootstrap();
    XMultiComponentFactory XMCF =
                       (XMultiComponentFactory)XCC.getServiceManager();
    XMultiServiceFactory XMSF1 =
                       (XMultiServiceFactory)XCC.getServiceManager();
    XComponentLoader XCL =
      (XComponentLoader)XMSF1.createInstance("com.sun.star.frame.Desktop");

    // open the spreadsheet file
    Console.Error.WriteLine("<opening "+pqssh+">");
    PropertyValue[] pPV = new PropertyValue[2];
    pPV[0] = new PropertyValue();
    pPV[0].Name = "Hidden";
    pPV[0].Value = new uno.Any(true);
    pPV[1] = new PropertyValue();
    pPV[1].Name = "ReadOnly";
    if (lreadonly) pPV[1].Value = new uno.Any(true);
    else pPV[1].Value = new uno.Any(false);
    string pqURL = "file:///"+pqssh.Replace("\\","/");
    XComponent XCo = XCL.loadComponentFromURL(pqURL,"_blank",0,pPV);
    XMultiServiceFactory XMSF2 = (XMultiServiceFactory)XCo;

    // get the spreadsheet sheet
    XSpreadsheets XSSs = ((XSpreadsheetDocument)XCo).getSheets();
    XNameAccess XNA = (XNameAccess)XSSs;
    XSpreadsheet XSS = null;
    try {
      XSS = (XSpreadsheet)XNA.getByName(pqsnm).Value;
    }
    catch (NoSuchElementException) {
      Console.WriteLine("Sheet name '"+pqsnm+"' not found.  Exiting.");
      return;
    }

/*  end: code to open a spreadsheet for reading and writing */
/*  begin: code to determine the last used spreadsheet row */

    // get the last used spreadsheet row (zero-based)
    XCellRange XCR = (XCellRange)XSS;
    XSheetCellCursor XSCC = XSS.createCursor();
    ((XUsedAreaCursor)XSCC).gotoEndOfUsedArea(true);
    XCellRangeAddressable XCRA = (XCellRangeAddressable)XSCC;
    CellRangeAddress CRA = XCRA.getRangeAddress();
    int nrlst = CRA.EndRow;

/*  end: code to determine the last used spreadsheet row */

    // read the spreadsheet contents into a List of items
    Console.Error.WriteLine("<reading spreadsheet rows>");
    List<Item> LIt = new List<Item>();
    for (int nrw = 2; nrw <= nrlst+1+1; nrw++) {

      // read and store the contents of a spreadsheet row
      Item Itm = new Item();
      Itm.nrow = nrw-1;

/*  begin: code to read cell contents as a string */

      string pqcell = "A"+nrw;
      XCell XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTextRange XTR = (XTextRange)XC;
      string pqval = XTR.getString();
      if (pqval == "") continue;
      Itm.pqitem = pqval.ToUpper();

/*  end: code to read cell contents as a string */

      pqcell = "B"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqbox = XTR.getString();

      pqcell = "C"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqdesc = XTR.getString();
      if (Itm.pqdesc == "") continue;

      pqcell = "D"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqJRC = XTR.getString();

      pqcell = "E"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqECH = XTR.getString();

      pqcell = "F"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqJCC = XTR.getString();

      pqcell = "G"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqval = XTR.getString();

      pqcell = "H"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqebay = XTR.getString();

      pqcell = "I"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqpic1 = XTR.getString();

      pqcell = "J"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqpic2 = XTR.getString();

      pqcell = "K"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqpic3 = XTR.getString();

      pqcell = "L"+nrw;
      XC = (XCell)XCR.getCellRangeByName(pqcell);
      XTR = (XTextRange)XC;
      Itm.pqpic4 = XTR.getString();

      if (lvrb) Console.WriteLine("    spreadsheet row "+(Itm.nrow+1)+
                                  " ("+Itm.pqitem+") read");

      // skip rows with duplicate item names
      bool ldup = false;
      foreach (Item Itmx in LIt) {
        if (Itm.pqitem == Itmx.pqitem) {
          ldup = true;
          Console.WriteLine("  duplicate item number '"+Itm.pqitem+
                            "' in row "+nrw+" (original in row "+
                            (Itmx.nrow+1)+") - skipped - delete it "+
                             "manually");
          break;
        }
      }
      if (ldup) continue;

      // add item to LIt
      LIt.Add(Itm);
    }

    Console.WriteLine("  "+LIt.Count+" spreadsheet rows found");

    // read and sort image file names from "images" folder
    Console.Error.WriteLine("<read and process image files from "+
                            pqimg+">");
    DirectoryInfo DI = new DirectoryInfo(pqimg);
    FileInfo[] pFI = DI.GetFiles("*.*",SearchOption.TopDirectoryOnly);
    List<string> Lpqflnm = new List<string>();
    foreach (FileInfo FI in pFI) Lpqflnm.Add(FI.Name);
    if (lvrb) foreach (FileInfo FI in pFI) {
      Console.WriteLine("    image file name '"+FI.Name+"'");
    }
    Lpqflnm.Sort();
    Console.WriteLine("  "+Lpqflnm.Count+" image file names found");

    // report and remove any bad image file names
    List<string> Lpqbdnm = new List<string>();
    for (int nnam = 0; nnam < Lpqflnm.Count; nnam++) {
      string pqimnm = Lpqflnm[nnam];

      // check name length
      if (pqimnm.Length != 7+4) {
        Console.WriteLine("  image file '"+pqimnm+"' has an invalid "+
                          "length of "+pqimnm.Length+" characters");
        Lpqbdnm.Add(pqimnm);
        continue;
      }

      // check name format
      lok = true;
      if (pqimnm[0] != 'm') lok = false;
      if (pqimnm[1] != 'd') lok = false;
      for (int nkar = 2; nkar <= 5; nkar++) {
        if (pqimnm[nkar] < '0' || pqimnm[nkar] > '9') lok = false;
      }
      if (pqimnm[7] != '.') lok = false;
      if (pqimnm[8] != 'j') lok = false;
      if (pqimnm[9] != 'p') lok = false;
      if (pqimnm[10] != 'g') lok = false;
      if (!lok) {
        Console.WriteLine("  image file '"+pqimnm+"' has an invalid "+
                          "name format");
        Lpqbdnm.Add(pqimnm);
        continue;
      }

      // check image column code
      if (pqimnm[6] < 'a' || pqimnm[6] > 'd') {
        Console.WriteLine("  image file '"+pqimnm+
                          "' has out-of-range image column code of '"+
                          pqimnm.Substring(6,1)+"'");
        Lpqbdnm.Add(pqimnm);
        continue;
      }

      // make sure there is a spreadsheet item row for the image file
      int nrssh = -1;
      foreach (Item Itm in LIt) {
        if (Itm.pqitem.ToLower() == pqimnm.Remove(6).ToLower()) {
          nrssh = Itm.nrow;
          break;
        }
      }
      if (nrssh < 0) {
        Console.WriteLine("  image file '"+pqimnm+"' belongs in "+
                          "a spreadsheet row that isn't defined");
        Lpqbdnm.Add(pqimnm);
        continue;
      }

      // remove the ".jpg" from the image name
      Lpqflnm[nnam] = pqimnm.Remove(7);
    }

    // remove bad image names from the list of image file names
    foreach (string pqbdnm in Lpqbdnm) Lpqflnm.Remove(pqbdnm);
    Console.WriteLine("  "+Lpqflnm.Count+" valid image file names");

/*  begin: code to get the DrawPage which contains shape images */

    // get DrawPage for this sheet
    XDrawPageSupplier XDPS = (XDrawPageSupplier)XSS;
    XDrawPage XDP = XDPS.getDrawPage();
    if (XDP == null) {
      Console.WriteLine("XDP is null.  Exiting.");
      return;
    }
    XShapes XSs = (XShapes)XDP;

/*  end: code to get the DrawPage which contains shape images */

    // if delete or refresh options are chosen, delete all shapes
    if (ldel || lrfr) {
      int ncnt = XSs.getCount();
      if (ncnt == 0) Console.Error.WriteLine("<no image shapes to delete>");
      if (ncnt == 1) Console.Error.WriteLine("<deleting 1 image shapes>");
      if (ncnt > 1) Console.Error.WriteLine("<deleting "+ncnt+
                                            " image shapes>");
      while (XSs.getCount() > 0) {
        XShape XSd = (XShape)XSs.getByIndex(0).Value;
        if (lvrb) Console.WriteLine("    deleting image shape "+
                    (ncnt-XSs.getCount())+" '"+((XNamed)XSd).getName()+"'");
        XDP.remove(XSd);
      }
      Console.WriteLine("  "+ncnt+" image shapes deleted");
      if (ldel && ncnt > 0) lchanged = true;
    }

    // skip the rest of the processing if delete option is selected
    if (!ldel) {

      // get all the graphic object shapes in the drawpage into LIm
      Console.Error.WriteLine("<reading "+XSs.getCount()+" shapes>");
      List<XShape> LXSdel = new List<XShape>();
      List<Img> LIm = new List<Img>();
      for (int ishp = 0; ishp < XSs.getCount(); ishp++) {
        Img Imx = new Img();

/*      begin: code to read an image shape */

        XShape XSdp = (XShape)XSs.getByIndex(ishp).Value;
        Imx.XS = XSdp;
        if (lvrb) Console.WriteLine("    reading shape "+ishp+
                    " '"+((XNamed)XSdp).getName()+"' from spreadsheet");

        // shape size and position
        Imx.Szs = XSdp.getSize();
        Imx.Pts = XSdp.getPosition();

        // get shape anchor
        XTextRange XTRsa = ((XTextContent)XSdp).getAnchor();

        // get cell address from anchor
        XCellAddressable XCAsa = (XCellAddressable)XTRsa;
        CellAddress CAsa = XCAsa.getCellAddress();
        Imx.nrow = CAsa.Row;
        Imx.ncol = CAsa.Column;

/*      end: code to read an image shape */

        // compute image base name from anchor row and column
        foreach (Item Itm in LIt) {
          if (Itm.nrow == Imx.nrow) {
            Imx.pqnam = Itm.pqitem.ToLower()+"abcd"[Imx.ncol-9];
            break;
          }
        }

        // compare image base name with shape name
        if (Imx.pqnam != ((XNamed)XSdp).getName()) {
          Console.WriteLine("  wrong shape name for shape at "+
                            (char)('A'+Imx.ncol)+(Imx.nrow+1));
          Console.WriteLine("  --name stored in shape is '"+
                            ((XNamed)XSdp).getName()+"'");
          Console.WriteLine("  --name from spreadsheet row item number "+
                            "is '"+Imx.pqnam+"'");
          continue;
        }

        // look through LIm for a duplicate of this name and mark for deletion
        bool ldup = false;
        foreach (Img Imy in LIm) {
          if (Imy.pqnam == Imx.pqnam) {
            ldup = true;
            LXSdel.Add(XSdp);
            Console.WriteLine("  duplicate shape named '"+Imx.pqnam+"'");
            break;
          }
        }
        if (ldup) continue;

/*      begin: code to read image graphic size & position from shape */

        // get graphic size (pixel size of JPG file) from shape
        XPropertySet XPSdp = (XPropertySet)XSdp;
        XGraphic XG = (XGraphic)XPSdp.getPropertyValue("Graphic").Value;
        Imx.Szg = (unoidl.com.sun.star.awt.Size)
                     ((XPropertySet)XG).getPropertyValue("SizePixel").Value;

/*      end: code to read image graphic size & position from shape */

        // mark graphic shape as present
        Imx.lshap = true;

        // look for image name in list of image files and mark it if found
        Imx.lfile = Lpqflnm.Contains(Imx.pqnam);

        // add the image to the list of image shapes
        LIm.Add(Imx);
      }

      // delete bad shapes found above
      if (LXSdel.Count > 0) {
        Console.WriteLine("<deleting "+LXSdel.Count+" bad shapes>");
        foreach (XShape XSdel in LXSdel) {
          Console.WriteLine("  shape '"+((XNamed)XSdel).getName()+
                            "' deleted");

/*        begin: code to delete an image shape */

          XDP.remove(XSdel);

/*        end: code to delete an image shape */

        }
        lchanged = true;
      }

      // only needed for a message below
      int nshps = LIm.Count;

      // add missing shapes to LIm for later processing
      int nmiss = 0;
      foreach (string pqimnm in Lpqflnm) {
        bool lfound = false;

        // ok if image file has a shape
        foreach (Img Imx in LIm) {
          if (Imx.pqnam == pqimnm) {
            lfound = true;
            break;
          }
        }

        // image file has no shape - add blank entry to LIm for it
        if (!lfound) {
          nmiss++;
          Img Imx = new Img();
          Imx.pqnam = pqimnm;
          Imx.lfile = true;
          Imx.lshap = false;
          if (lvrb) Console.WriteLine("    image file '"+pqimnm+
                                      "' has no shape");
          LIm.Add(Imx);
        }
      }

      // add any missing image shapes
      if (nmiss == 0) Console.Error.WriteLine("<no new shapes to add>");
      else if (nmiss == 1) Console.Error.WriteLine("<adding 1 new shape>");
      else Console.Error.WriteLine("<adding "+nmiss+" new shapes>");
      if (lvrb) Console.WriteLine("    starting with "+nshps+" shapes and "+
                                  Lpqflnm.Count+" image files");
      if (nmiss > 0) {

/*      begin: code to set up to insert a new image shape */

        // get a graphic provider
        // (used for creating graphic objects which contain images)
        object oGP = XMCF.createInstanceWithContext(
                                  "com.sun.star.graphic.GraphicProvider",XCC);
        if (oGP == null) {
          Console.WriteLine("oGP is null.  Aborting.");
          return;
        }
        XGraphicProvider XGP = (XGraphicProvider)oGP;
        if (XGP == null) {
          Console.WriteLine("XGP is null.  Aborting.");
          return;
        }

/*      end: code to set up to insert a new image shape */

        // add missing shapes for new image files
        for (int nshp = 0; nshp < LIm.Count; nshp++) {
          Img Imx = LIm[nshp];
          if (Imx.lfile && !Imx.lshap) {

            // get row and column for image
            Imx.nrow = -1;
            foreach (Item Itm in LIt) {
              if (Itm.pqitem.ToLower() == Imx.pqnam.Remove(6).ToLower()) {
                Imx.nrow = Itm.nrow;
                break;
              }
            }

            // if no item row is found, use row 0 to avoid exceptions
            if (Imx.nrow < 0) {
              Console.WriteLine("  no item row for image '"+
                                Imx.pqnam+"' being added");
              Console.WriteLine("  --putting it in first row of "+
                                "spreadsheet");
              Imx.nrow = 0;
              continue;
            }
            Imx.ncol = 10+Imx.pqnam[6]-'a'-1;

            // make shape for picture to be added
            if (lvrb) Console.WriteLine("    creating and adding new shape"+
                                        " for image '"+Imx.pqnam+"'");

/*          begin: code to insert an image shape */

            // create graphic object containing image
            pPV = new PropertyValue[1];
            pPV[0] = new PropertyValue();
            pPV[0].Name = "URL";
            pqURL = "file:///"+pqimg.Replace("\\","/")+Imx.pqnam+".jpg";
            pPV[0].Value = new uno.Any(pqURL);
            XGraphic XG = XGP.queryGraphic(pPV);
            if (XG == null) {
              Console.WriteLine("XG is null.  Aborting.");
              return;
            }

            // create empty graphic shape object
            XShape XS = (XShape)XMSF2.createInstance(
                            "com.sun.star.drawing.GraphicObjectShape");
            if (XS == null) {
              Console.WriteLine("XS is null.  Aborting.");
              return;
            }

            // add shape to the cell
            try {
              XSs.add(XS);
            }
            catch (System.Exception E) {
              Console.WriteLine("Adding '"+Imx.pqnam+"': "+E);
            }
            lchanged = true;

            // set XShape XGraphic containing image
            XPropertySet XPS = (XPropertySet)XS;
            XPS.setPropertyValue("Name",new uno.Any(Imx.pqnam));
            XPS.setPropertyValue("Graphic",new uno.Any(typeof(XGraphic),XG));

            // add anchor to the shape
            XCell XC = (XCell)XCR.getCellByPosition(Imx.ncol,Imx.nrow);
            XPS.setPropertyValue("Anchor",new uno.Any(typeof(XCell),XC));

/*          end: code to insert an image shape */

            // find index of new shape and save in Imx(->LIm)
            for (int ishp = 0; ishp < XSs.getCount(); ishp++) {
              XShape XSdp = (XShape)XSs.getByIndex(ishp).Value;
              XNamed XNdp = (XNamed)XSdp;
              if (XNdp.getName() == Imx.pqnam) {
                Imx.XS = XSdp;
                break;
              }
            }

            // read the image file size and put in Szg
            Image I = Image.FromFile(pqimg+Imx.pqnam+".jpg");
            Imx.Szg.Height = I.Height;
            Imx.Szg.Width = I.Width;

            // fill in more stuff in LIm entry
            Imx.Szs = new unoidl.com.sun.star.awt.Size();
            Imx.Pts = new unoidl.com.sun.star.awt.Point();

            // arbitrary size and location for now so they won't be undefined
            Imx.Szs.Height = 2500;
            Imx.Szs.Width = 2500;
            Imx.Pts.X = 100;
            Imx.Pts.Y = 100;

            // write changed LIm entry back out
            LIm[nshp] = Imx;
          }
        }
      }

      // remove shapes that don't have image files
      int ndel = 0;
      foreach (Img Imx in LIm) if (!Imx.lfile) ndel++;
      if (ndel == 1) {
        Console.Error.WriteLine("<removing 1 shape without an image file>");
      }
      if (ndel > 1) {
        Console.Error.WriteLine("<removing "+ndel+
                                " shapes without image files>");
      }
      for (int nimg = 0; nimg < LIm.Count; nimg++) {
        if (!LIm[nimg].lfile) {
          if (lvrb) Console.WriteLine("    removing shape '"+
                                      ((XNamed)LIm[nimg].XS).getName()+"'");
          XDP.remove(LIm[nimg].XS);
          LIm.RemoveAt(nimg);
          nimg--;
          lchanged = true;
        }
      }
      if (ndel == 1) Console.WriteLine("  1 shape removed");
      if (ndel > 1) Console.WriteLine("  "+ndel+" shapes removed");

      // scale and position each shape to fit and center in cell
      Console.Error.WriteLine("<checking scale/position/name of "+
                              LIm.Count+" shape images>");
      int nchg = 0;
      foreach (Img Imx in LIm) {
        bool lchg = false;
        if (lvrb) Console.Write("    image '"+Imx.pqnam+"'");

        // read the JPG file into an Image instance
        Image I = Image.FromFile(pqimg+Imx.pqnam+".jpg");

/*      begin: code to read the size and position from a cell */

        // get cell size and position
        XCell XC = (XCell)XCR.getCellByPosition(Imx.ncol,Imx.nrow);
        Imx.Szc = (unoidl.com.sun.star.awt.Size)
                  (((XPropertySet)XC).getPropertyValue("Size").Value);
        Imx.Ptc = (unoidl.com.sun.star.awt.Point)
                  (((XPropertySet)XC).getPropertyValue("Position").Value);

/*      end: code to read the size and position from a cell */

        // compute scaled size from cell size and image size
        double sclx = (double)Imx.Szc.Width/(double)Imx.Szg.Width;
        double scly = (double)Imx.Szc.Height/(double)Imx.Szg.Height;
        if (sclx < scly) {
          Imx.Szn.Width = (int)(sclx*Imx.Szg.Width);
          Imx.Szn.Height = (int)(sclx*Imx.Szg.Height);
        }
        else {
          Imx.Szn.Width = (int)(scly*Imx.Szg.Width);
          Imx.Szn.Height = (int)(scly*Imx.Szg.Height);
        }

/*      begin: code to write the size, position, and name into a cell */

        // write scaled size into shape if it is different
        if (Imx.Szs.Width != Imx.Szn.Width ||
            Imx.Szs.Height != Imx.Szn.Height) {
          if (lvrb) Console.Write(" size ("+Imx.Szs.Width+","+
                                  Imx.Szs.Height+")->("+Imx.Szn.Width+","+
                                  Imx.Szn.Height+")");
          Imx.XS.setSize(Imx.Szn);
          lchg = true;
          lchanged = true;
        }

        // adjust indent of image in cell
        Imx.Ptn.X = Imx.Ptc.X;
        Imx.Ptn.Y = Imx.Ptc.Y;
        if (Imx.Szn.Width < Imx.Szc.Width)
                     Imx.Ptn.X = Imx.Ptc.X+(Imx.Szc.Width-Imx.Szn.Width)/2;
        if (Imx.Szn.Height < Imx.Szc.Height)
                     Imx.Ptn.Y = Imx.Ptc.Y+(Imx.Szc.Height-Imx.Szn.Height)/2;

        // write adjusted indent into shape if it is different
        if (Imx.Pts.X != Imx.Ptn.X || Imx.Pts.Y != Imx.Ptn.Y) {
          if (lvrb) Console.Write(" position ("+Imx.Pts.X+","+
                                  Imx.Pts.Y+")->("+Imx.Ptn.X+","+
                                  Imx.Ptn.Y+")");
          Imx.XS.setPosition(Imx.Ptn);
          lchg = true;
          lchanged = true;
        }

        // write image name into shape if needed
        if (((XNamed)Imx.XS).getName() != Imx.pqnam) {
          ((XNamed)Imx.XS).setName(Imx.pqnam);

/*        end: code to write the size, position, and name into a cell */

          lchg = true;
          lchanged = true;
        }
        if (lchg) nchg++;
        if (lvrb) {
          if (lchg) Console.WriteLine();
          else Console.WriteLine(" unchanged");
        }
      }
      if (nchg == 0) Console.WriteLine("  no shape images updated");
      if (nchg == 1) Console.WriteLine("  1 shape image updated");
      if (nchg > 1) Console.WriteLine("  "+nchg+" shape images updated");

      // Check that shape's image file X/Y size matches actual image file
      // size.  If not, image file has changed and shape image must be
      // updated.
      //
      // For now, this needs to be handled manually but a warning would
      // be nice.  This isn't a good check anyhow because many (most)
      // image edits don't change the image dimensions.
      //
      // The right way to do this is to get the image file stored in the
      // shape and compare it to the image file on disk.  This can be done
      // because of the "Edit using external tool" function in Calc, but
      // I don't know how and for me it's not worth the research project
      // to learn how.  Plus it would be quite slow, and the -rfr option
      // will take care of it.
    }

/*  begin: code to save and close the spreadsheet and close the desktop */

    // save the spreadsheet if it is changeable and changed
    XModifiable XM = (XModifiable)XCo;
    XM.setModified(!lreadonly);
    if (!lreadonly && lchanged) {
      Console.Error.WriteLine("<saving>");
      XStorable XSt = (XStorable)XCo;
      XSt.store();
    }
    if (lreadonly && lchanged) {
      Console.Error.WriteLine("<changes not saved>");
    }

    // close the spreadsheet
    Console.Error.WriteLine("<closing>");
    XCloseable XCl = (XCloseable)XCo;
    XCl.close(true);
    Thread.Sleep(500); // needed? probably not, but who knows

    // close the desktop if nothing else has it open
    XDesktop XD = (XDesktop)XCL;
    XEnumeration XE = XD.getComponents().createEnumeration();
    if (XE.hasMoreElements()) {
      Console.WriteLine("Process not terminated: "+
                        "another instance is running");
    }
    else {
      lok = XD.terminate();
      if (lok) Console.WriteLine("Process terminated successfully");
      else Console.WriteLine("Process termination failed"+
                             " (bug in LibreOffice?)");
    }

/*  end: code to save and close the spreadsheet and close the desktop */

  }
}
